from collections import OrderedDict
import json
import time

import psycopg2

from keyrock_core import json_util

from . import util
from . import exc
from .connection import Connection

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def wait_for_server(config, timeout_sec=30, interval_sec=5):
    logger.debug(f'wait_for_server timeout:{timeout_sec} interval: {interval_sec}')

    assert(interval_sec >= 0)
    assert(timeout_sec >= 0)

    waited_sec = 0
    try_count = 0
    while True:
        try_count += 1
        if is_server_ready(config):
            logger.debug(f'wait_for_server succeeded after {waited_sec} seconds and {try_count} tries')
            return True
        else:
            if waited_sec > timeout_sec:
                raise Exception(f'server is not ready after {waited_sec} seconds and {try_count} tries')
            else:
                logger.debug(f'wait_for_server will retry in {interval_sec} seconds')
                time.sleep(interval_sec)
                waited_sec += interval_sec
    raise


def is_server_ready(config):
    config = Connection.ConfigSchema().load(config)
    logger.debug(f'is_server_ready')

    # Create a temp config for the root "postgres" database
    root_config = config.copy()
    root_config['db'] = 'postgres'

    try:
        conn = Connection(config)
    except exc.DatabaseDoesNotExist as e:
        logger.info(f'server is ready: {str(e)}')
        return True
    except Exception as e:
        logger.info(f'server is not ready: {str(e)}')
        return False

    logger.info(f'server is ready')
    return True


def database_exists(config):
    try:
        conn = Connection(config)
    except exc.DatabaseDoesNotExist:
        return False
    return True


def create_database(config):
    config = Connection.ConfigSchema().load(config)

    db_name = config['db']
    create_cmd = f'CREATE DATABASE {db_name}'

    # Create a temp config for the root "postgres" database
    root_config = config.copy()
    root_config['db'] = 'postgres'

    conn = Connection(root_config, autocommit=True)
    conn.cmd(create_cmd)


def drop_database(config):
    config = Connection.ConfigSchema().load(config)

    db_name = config['db']
    drop_cmd = f'DROP DATABASE {db_name}'

    # Create a temp config for the root "postgres" database
    root_config = config.copy()
    root_config['db'] = 'postgres'

    conn = Connection(root_config, autocommit=True)
    conn.cmd(drop_cmd)


def get_structure(conn):
    db_info = conn.get_single_result(
        "SELECT"
        " current_catalog AS catalog,"
        " current_user AS user"
        #" current_schemas(FALSE) AS schemas,"
    )
    logger.debug('DB INFO:')
    logger.debug(json.dumps(db_info, indent=2))

    # Catalog is the root element (can be only one per connection)
    catalog = {}
    catalog['schemas'] = get_schemas(conn)

    return catalog


def get_schemas(conn):
    schemas = {}
    schema_query = (
        "SELECT * FROM information_schema.schemata"
        " WHERE TRUE"
        #f" AND schema_owner = '{db_info['user']}'"
        # Exclude system schemas:
        " AND schema_name NOT LIKE 'pg_%'"
        " AND schema_name != 'information_schema'"
        " ORDER BY schema_name ASC"
    )
    schema_list = conn.get_list_result(schema_query, 'schema_name')

    for schema_name in schema_list:
        schema = schemas.setdefault(schema_name, {})
        schema['tables'] = get_tables(conn, schema_name)

    return schemas


def get_tables(conn, schema_name):
    tables = {}
    table_query = (
        "SELECT * FROM information_schema.tables"
        f" WHERE table_schema = '{schema_name}'"
        " ORDER BY table_name ASC"
    )
    table_list = conn.get_list_result(table_query, 'table_name')

    for table_name in table_list:
        table = tables.setdefault(table_name, {})
        table['columns'] = get_columns(conn, schema_name, table_name)
        table['constraints'] = get_constraints(conn, schema_name, table_name)

    return tables


def get_columns(conn, schema_name, table_name):
    columns = {}
    columns_query = (
        "SELECT"
        "    c.column_name, c.data_type, c.character_maximum_length, c.is_nullable"
        "  , c.column_default, c.ordinal_position"
        "  , c.is_identity, c.identity_start, c.identity_increment"
        "  , e.data_type AS element_type"
        " FROM information_schema.columns c"
        " LEFT JOIN information_schema.element_types e"
        "  ON (c.table_catalog, c.table_schema, c.table_name, 'TABLE', c.dtd_identifier)"
        "   = (e.object_catalog, e.object_schema, e.object_name, e.object_type, e.collection_type_identifier)"    
        f" WHERE c.table_schema = '{schema_name}'"
        f" AND c.table_name = '{table_name}'"
    )
    column_list = conn.get_list_result(columns_query)

    for column_src in column_list:
        column_name = column_src['column_name']
        columns[column_name] = {
            'is_identity': (column_src['is_identity'] == 'YES'),
            'identity_start': column_src['identity_start'],
            'identity_increment': column_src['identity_increment'],
            'type': column_src['data_type'].upper(),
            'size': column_src['character_maximum_length'],
            'allow_null': (column_src['is_nullable'] == 'YES'),
            'default': util.convert_default(column_src['column_default']),
            'pos': column_src['ordinal_position'],
            'element_type': column_src['element_type'],
        }

    return columns


def get_constraints(conn, schema_name, table_name):
    constraints = {}
    constraints_query = (
        "SELECT"
        "    constraint_name, constraint_type"
        " FROM information_schema.table_constraints"
        f" WHERE table_schema = '{schema_name}'"
        f" AND table_name = '{table_name}'"
    )
    constraint_list = conn.get_list_result(constraints_query)
    #print('CONSTRAINTS', json.dumps(constraint_list, indent=2))

    for constraint_src in constraint_list:
        constraint_name = constraint_src['constraint_name']
        constraint_type = constraint_src['constraint_type'].upper()

        columns_query = (
            "SELECT"
            "    table_name, column_name"
            " FROM information_schema.key_column_usage"
            f" WHERE table_schema = '{schema_name}'"
            f" AND constraint_name = '{constraint_name}'"
        )
        column_list = conn.get_list_result(columns_query)
        #print('CONSTRAINT', constraint_name, constraint_type)
        #print('CONSTRAINT COLUMNS', json.dumps(column_list, indent=2))

        if constraint_type == 'FOREIGN KEY':
            foreign_keys = constraints.setdefault('foreign', {})
            (local_col, foreign_path) = get_foreign_constraint(conn, schema_name, table_name, constraint_name)
            foreign_keys[local_col] = foreign_path
        elif constraint_type == 'PRIMARY KEY':
            constraints['primary'] = [ c['column_name'] for c in column_list ]
        elif constraint_type == 'UNIQUE':
            # Strip out table_name prefix on unique keys
            prefix = f'{table_name}_'
            if constraint_name.startswith(prefix):
                constraint_name = constraint_name[len(prefix):]
            unique_keys = constraints.setdefault('unique', {})
            unique_keys[constraint_name] = [ c['column_name'] for c in column_list ]
        elif constraint_type == 'CHECK':
            # not-null check
            pass
        else:
            logger.warning(f"Unhandled constraint type: '{constraint_type}' for '{constraint_name}")

    return constraints


def get_foreign_constraint(conn, schema_name, table_name, constraint_name):
    columns_query = (
        "SELECT"
        "    cu.table_name AS foreign_table_name,"
        "    cu.column_name AS foreign_column_name,"
        "    ku.table_name AS local_table_name,"
        "    ku.column_name AS local_column_name"
        " FROM information_schema.constraint_column_usage cu"
        " LEFT JOIN information_schema.key_column_usage ku"
        "  ON (cu.table_schema, cu.constraint_name)"
        "   = (ku.table_schema, ku.constraint_name)"
        f" WHERE cu.table_schema = '{schema_name}'"
        f" AND cu.constraint_name = '{constraint_name}'"
    )
    column_map = conn.get_single_result(columns_query)

    local_column_name = column_map['local_column_name']
    foreign_target = f"{column_map['foreign_table_name']}.{column_map['foreign_column_name']}"

    return (local_column_name, foreign_target)



def config_to_str(config):
    try:
        return f"{config['host']}:{config['port']}.{config['db']}"
    except:
        return '[invalid config]'


class DatabaseDoesNotExist(Exception):
    def __init__(self, config):
        config_str = config_to_str(config)
        msg = f'Database does not exist: {config_str}'
        super().__init__(msg)



import unittest

import os
import json

try:
    import psycopg2
except:
    psycopg2 = None

from keyrock_core import config_loader
from keyrock_core import json_util

from . import *

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), '.'))


class TestMigration(unittest.TestCase):
    config = {
        'host': 'pg',
        'port': 5432,
        'db': 'migration_test',
        'user': 'postgres',
        'pass': 'postgres',
    }

    @classmethod
    def setUpClass(cls):
        if psycopg2 is None:
            raise unittest.SkipTest('psycopg2 not installed')

        wait_for_server(cls.config, timeout_sec=15)

        if database_exists(cls.config):
            logger.warning(f'Test database already exists: {cls.config}')
            drop_database(cls.config)

        try:
            create_database(cls.config)
        except Exception as e:
            cls.tearDownClass()
            raise e

    @classmethod
    def tearDownClass(cls):
        assert (database_exists(cls.config)), 'database does not exist'

        # Remove the test database
        drop_database(cls.config)

    def setUp(self):
        self.assertTrue(database_exists(self.config), 'database does not exist')
        self.conn = Connection(self.config)

    def test_root(self):
        try:
            self.run_migration('test_data/unit_test_init.yml')
            self.run_migration('test_data/unit_test_2.yml')
            # @todo: Load some test data and run more migrations
        except Exception as e:
            logger.error(str(e))
            self.conn.disconnect()
            raise e

    def run_migration(self, target_config_file):
        logger.debug(f'run migration: {target_config_file}')

        current_structure = get_structure(self.conn)
        #logger.debug('current_structure')
        #logger.debug(json.dumps(current_structure, indent=2))

        target_structure = config_loader.load(os.path.join(ROOT_DIR, target_config_file))
        #logger.debug('target_structure:')
        #logger.debug(json.dumps(target_structure, indent=2))

        structure_diff = catalog_diff(current_structure, target_structure)
        logger.debug('structure_diff (changes only):')
        logger.debug(json.dumps(json_util.strip_unchanged(structure_diff), indent=2))

        cmd_list = diff_to_cmd_list(structure_diff)
        logger.debug('cmd_list:')
        logger.debug(json.dumps(cmd_list, indent=2))

        error_list = self.conn.exec_cmd_list(cmd_list)
        logger.debug('error_list:')
        logger.debug(json.dumps(error_list, indent=2))
        self.assertEqual(len(error_list), 0)

        #
        # Verify that the new structure matches the target
        #
        updated_structure = get_structure(self.conn)
        verify_diff = catalog_diff(updated_structure, target_structure)
        logger.debug('verify_diff (should have no changes):')
        logger.debug(json.dumps(json_util.strip_unchanged(verify_diff), indent=2))
        self.assertEqual(verify_diff['op'], None)

        cmd_list = diff_to_cmd_list(verify_diff)
        logger.debug('re-apply cmd list (should be empty):')
        logger.debug(json.dumps(cmd_list, indent=2))
        self.assertEqual(len(cmd_list), 0)


class TestPsql(unittest.TestCase):
    config = {
        'host': 'pg',
        'port': 5432,
        'db': 'operations_test',
        'schema': 'test_schema',
        'user': 'postgres',
        'pass': 'postgres',
    }

    target_config_file = 'test_data/op_test_init.yml'

    @classmethod
    def setUpClass(cls):
        if psycopg2 is None:
            raise unittest.SkipTest('psycopg2 not installed')

        wait_for_server(cls.config, timeout_sec=15)

        if database_exists(cls.config):
            logger.warning(f'Test database already exists: {cls.config}')
            drop_database(cls.config)

        #
        # Set up the initial structure
        #
        try:
            create_database(cls.config)
            conn = Connection(cls.config)
            current_structure = get_structure(conn)
            target_structure = config_loader.load(os.path.join(ROOT_DIR, cls.target_config_file))
            structure_diff = catalog_diff(current_structure, target_structure)
            cmd_list = diff_to_cmd_list(structure_diff)
            error_list = conn.exec_cmd_list(cmd_list)
        except Exception as e:
            cls.tearDownClass()
            raise e

    @classmethod
    def tearDownClass(cls):
        assert (database_exists(cls.config)), 'database does not exist'

        # Remove the test database
        drop_database(cls.config)

    def setUp(self):
        self.assertTrue(database_exists(self.config), 'database does not exist')
        self.conn = Connection(self.config, autocommit=True)

    def tearDown(self):
        if self.conn:
            self.conn.disconnect()

    def test_basic_ops(self):
        self.conn.cmd("DELETE FROM test_table_a")

        for test_val in [11, 12, 13]:
            row_id = self.conn.insert('test_table_a', {'int_column': test_val})
            query_str = f"SELECT int_column FROM test_table_a WHERE {util.sql_eq('id', row_id)}"
            result = self.conn.get_single_result(query_str, 'int_column')
            self.assertEqual(result, test_val)

            test_id = self.conn.update('test_table_a', {'int_column': test_val+1}, row_id)
            self.assertEqual(test_id, row_id)

            result = self.conn.get_value('test_table_a', row_id)
            self.assertEqual(result['int_column'], test_val+1)

            result = self.conn.get_value('test_table_a', row_id, ['int_column'])
            self.assertEqual(result['int_column'], test_val+1)

            result = self.conn.get_value('test_table_a', row_id, 'int_column')
            self.assertEqual(result, test_val+1)

        query_str = "SELECT * FROM test_table_a LIMIT 2"
        result = self.conn.get_list_result(query_str)
        self.assertEqual(len(result), 2)

        test_id = self.conn.insert('test_table_a', {'int_column': 69})
        delete_count = self.conn.delete('test_table_a', test_id)
        self.assertEqual(delete_count, 1)

        delete_count = self.conn.delete('test_table_a', None, key='str_column_opt')
        query_str = "SELECT COUNT(1) AS row_count FROM test_table_a"
        result = self.conn.get_single_result(query_str)
        self.assertEqual(result['row_count'], 0)

    def test_upsert(self):
        self.conn.cmd("DELETE FROM test_table_b")

        test_id = self.conn.upsert('test_table_b', {'str_column': 'abc'}, id='testkeyvalue', key='test_key')
        self.assertEqual(test_id, 'testkeyvalue')

        test_id = self.conn.upsert('test_table_b', {'str_column': 'def', 'test_key': 'testkeyvalue'}, key='test_key')
        self.assertEqual(test_id, 'testkeyvalue')

        test_id = self.conn.upsert('test_table_b', {'str_column': 'ghi', 'test_key': 'newkeyvalue'}, id='testkeyvalue', key='test_key')
        self.assertEqual(test_id, 'newkeyvalue')

        test_id = self.conn.upsert('test_table_b', {'str_column': 'new', 'test_key': 'insertkeyvalue'}, id='testkeyvalue', key='test_key')
        self.assertEqual(test_id, 'insertkeyvalue')

        id_list = ['insertkeyvalue', 'newkeyvalue']
        delete_count = self.conn.delete('test_table_b', id_list, key='test_key')
        self.assertEqual(delete_count, 2)

    def test_insert_list(self):
        self.conn.cmd("DELETE FROM test_table_a")

        item_list = [
            {'int_column': 1},
            {'int_column': 2},
            {'int_column': 3},
        ]
        id_list = self.conn.insert_list('test_table_a', item_list)
        self.assertEqual(len(id_list), 3)

        delete_count = self.conn.delete('test_table_a', id_list)
        self.assertEqual(delete_count, 3)



import json

from keyrock_core import json_util

from . import exc
from . import util

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

def diff_to_cmd_list(node, cmd_list=None, context=None):
    if cmd_list is None:
        cmd_list = []

    if context is None:
        context = {}

    if node['op'] is None:
        return cmd_list

    if node['op'] == 'dict':
        for group_key, group_node in node['diff'].items():
            cmd_gen = CmdGen.get_cmd_gen(group_key)
            if cmd_gen is not None:
                cmd_gen.parse_diff(group_node, cmd_list, context)
    return cmd_list


class CmdGen():
    group_map = {}

    @staticmethod
    def register_cmd_gen(group_key, context_key, cmd_gen_class):
        CmdGen.group_map[group_key] = cmd_gen_class(context_key)

    @staticmethod
    def get_cmd_gen(group_key):
        cmd_gen_instance = CmdGen.group_map.get(group_key)
        if cmd_gen_instance is None:
            logger.debug(f'cmd_gen_class not registered for: {group_key}')
            return None
        return cmd_gen_instance

    def __init__(self, context_key):
        self.context_key = context_key

    def parse_diff(self, group_node, cmd_list, context):
        group_op = group_node['op']
        if group_op == 'dict':
            for item_key, item_node in group_node['diff'].items():
                self.write(item_key, item_node, cmd_list, context)
        elif group_op == 'add':
            raise NotImplementedError('group add implemented')
        elif group_op == 'del':
            raise NotImplementedError('group del not implemented')

    def update_context(self, item_key, context):
        context[self.context_key] = item_key

    def write(self, item_key, item_node, cmd_list, context):
        op = item_node['op']
        if op == 'dict':
            return self.write_dict(item_key, item_node['diff'], cmd_list, context)
        if op == 'add':
            return self.write_add(item_key, item_node['val'], cmd_list, context)
        if op == 'del':
            return self.write_del(item_key, item_node['val_from'], cmd_list, context)
        if op == 'mod':
            return self.write_mod(item_key, item_node['val_from'], item_node['val'], cmd_list, context)
        if op is None:
            return None
        assert(False), f"unrecognized op: {op}"

    def write_dict(self, item_key, item_node, cmd_list, context):
        logger.debug(f"write_dict for {self.context_key}: {item_key}")
        self.update_context(item_key, context)
        for group_key, group_node in item_node.items():
            cmd_gen = CmdGen.get_cmd_gen(group_key)
            if cmd_gen is not None:
                cmd_gen.parse_diff(group_node, cmd_list, context)

    def write_mod(self, item_key, item_data_from, item_data, cmd_list, context):
        logger.debug(f"write_mod for {self.context_key}: {item_key} {item_data_from} -> {item_data}")
        self.update_context(item_key, context)

    def write_add(self, item_key, item_data, cmd_list, context):
        logger.debug(f"write_add for {self.context_key}: {item_key} {item_data}")
        self.update_context(item_key, context)

    def add_child_data(self, group_key, item_data, cmd_list, context):
        if item_data is not None:
            if group_key in item_data:
                group_data = item_data[group_key]
                cmd_gen = CmdGen.get_cmd_gen(group_key)
                for sub_item_key, sub_item_data in group_data.items():
                    cmd_gen.write_add(sub_item_key, sub_item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        logger.debug(f"write_del for {self.context_key}")


class CatalogCmdGen(CmdGen):
    pass

CmdGen.register_cmd_gen('catalogs', 'catalog', CatalogCmdGen)


class SchemaCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        cmd_list.append(f"CREATE SCHEMA {item_key} AUTHORIZATION CURRENT_USER")
        super().write_add(item_key, item_data, cmd_list, context)
        self.add_child_data('tables', item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        cmd_list.append(f"DROP SCHEMA {item_key}")
        super().write_del(item_key, item_data, cmd_list, context)

CmdGen.register_cmd_gen('schemas', 'schema', SchemaCmdGen)


class TableCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        cmd_list.append(f"CREATE TABLE {context['schema']}.{item_key} ()")
        super().write_add(item_key, item_data, cmd_list, context)
        self.add_child_data('columns', item_data, cmd_list, context)
        self.add_child_data('constraints', item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        cmd_list.append(f"DROP TABLE {context['schema']}.{item_key}")
        super().write_del(item_key, item_data, cmd_list, context)

CmdGen.register_cmd_gen('tables', 'table', TableCmdGen)


class ColumnCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" ADD COLUMN {item_key} {item_data['type']}"
        )

        if item_data.get('size'):
            cmd_str += f"({item_data['size']})"

        if item_data.get('is_identity'):
            identity_start = item_data.get('identity_start', 1)
            identity_method = item_data.get('identity_method', 'BY DEFAULT')
            cmd_str += f' GENERATED {identity_method} AS IDENTITY (START WITH {identity_start})'

        if not item_data.get('allow_null'):
            cmd_str += ' NOT NULL'

        if 'default' in item_data:
            default_val = util.to_sql(item_data['default'])
            cmd_str += f' DEFAULT {default_val}'
        cmd_list.append(cmd_str)
        super().write_add(item_key, item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        cmd_list.append(f"ALTER TABLE {context['schema']}.{context['table']} DROP COLUMN {item_key}")
        super().write_del(item_key, item_data, cmd_list, context)

    def write_dict(self, item_key, item_node, cmd_list, context):
        item_data = {}
        for prop_key, prop_node in item_node.items():
            if prop_node['op'] != 'del':
                prop_data = json_util.assemble_node_data(prop_node)
                item_data[prop_key] = prop_data

        # Update type (and/or size)
        type_op = json_util.get_property_op(item_node, 'type')
        size_op = json_util.get_property_op(item_node, 'size')
        if type_op is not None or size_op is not None:
            self.update_type(item_key, item_data, cmd_list, context)

        # Update identity
        # @todo: this could be cleaned up
        is_identity_op = json_util.get_property_op(item_node, 'is_identity')
        identity_start_op = json_util.get_property_op(item_node, 'identity_start')
        identity_method_op = json_util.get_property_op(item_node, 'identity_method')
        if is_identity_op is not None or identity_start_op is not None or identity_method_op is not None:
            is_identity = item_data.get('is_identity')
            if is_identity:
                identity_start = item_data.get('identity_start', 1)
                identity_method = item_data.get('identity_method', 'BY DEFAULT')
                if is_identity_op == 'mod' or is_identity_op == 'add':
                    cmd_str = (
                        f"ALTER TABLE {context['schema']}.{context['table']}"
                        f" ALTER COLUMN {item_key} ADD GENERATED {identity_method} AS IDENTITY (START WITH {identity_start})"
                    )
                    cmd_list.append(cmd_str)
                elif identity_start_op is not None:
                    cmd_str = (
                        f"ALTER TABLE {context['schema']}.{context['table']}"
                        f" ALTER COLUMN {item_key} SET GENERATED {identity_method} RESTART WITH {identity_start}"
                    )
                    cmd_list.append(cmd_str)
            else:
                cmd_str = (
                    f"ALTER TABLE {context['schema']}.{context['table']}"
                    f" ALTER COLUMN {item_key} DROP IDENTITY IF EXISTS"
                )
                cmd_list.append(cmd_str)

        # Change allow null
        allow_null_op = json_util.get_property_op(item_node, 'allow_null')
        if allow_null_op is not None:
            self.update_allow_null(item_key, item_data, cmd_list, context)

        # Change default value
        # @todo: this could be cleaned up
        default_val_op = json_util.get_property_op(item_node, 'default')
        if default_val_op == 'del':
            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" ALTER COLUMN {item_key} DROP DEFAULT"
            )
            cmd_list.append(cmd_str)
        elif default_val_op is not None:
            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" ALTER COLUMN {item_key} SET DEFAULT {item_data['default']}"
            )
            cmd_list.append(cmd_str)

        super().write_dict(item_key, item_node, cmd_list, context)

    def update_type(self, item_key, item_data, cmd_list, context):
        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" ALTER COLUMN {item_key} TYPE {item_data['type']}"
        )
        if item_data.get('size'):
            cmd_str += f"({item_data['size']})"
        cmd_list.append(cmd_str)

    def update_allow_null(self, item_key, item_data, cmd_list, context):
        allow_null = item_data.get('allow_null')
        if allow_null:
            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" ALTER COLUMN {item_key} DROP NOT NULL"
            )
            cmd_list.append(cmd_str)
        else:
            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" ALTER COLUMN {item_key} SET NOT NULL"
            )
            cmd_list.append(cmd_str)

CmdGen.register_cmd_gen('columns', 'column', ColumnCmdGen)


class ConstraintCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        if item_key == 'primary':
            if not isinstance(item_data, list):
                item_data = [item_data]
            key_str = ', '.join(item_data)

            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" ADD PRIMARY KEY({key_str})"
            )
            cmd_list.append(cmd_str)
        else:
            cmd_gen = CmdGen.get_cmd_gen(item_key)
            for sub_key, sub_data in item_data.items():
                cmd_gen.write_add(sub_key, sub_data, cmd_list, context)

        super().write_add(item_key, item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        if item_key == 'primary':
            cmd_str = (
                f"ALTER TABLE {context['schema']}.{context['table']}"
                f" DROP CONSTRAINT IF EXISTS {context['table']}_pkey"
            )
            cmd_list.append(cmd_str)
        else:
            cmd_gen = CmdGen.get_cmd_gen(item_key)
            # Drop each of the constraints
            for sub_key, sub_data in item_node.items():
                cmd_gen.write_del(sub_key, sub_data, cmd_list, context)

        super().write_del(item_key, item_data, cmd_list, context)

    def write_mod(self, item_key, item_data_from, item_data, cmd_list, context):
        if item_key == 'primary':
            self.write_del(item_key, item_data_from, cmd_list, context)
            self.write_add(item_key, item_data, cmd_list, context)
        super().write_mod(item_key, item_data_from, item_data, cmd_list, context)

    def write_dict(self, item_key, item_node, cmd_list, context):
        if item_key == 'primary':
            pass
        else:
            cmd_gen = CmdGen.get_cmd_gen(item_key)
            for sub_key, sub_node in item_node.items():
                cmd_gen.write(sub_key, sub_node, cmd_list, context)

        super().write_dict(item_key, item_node, cmd_list, context)

CmdGen.register_cmd_gen('constraints', 'constraint', ConstraintCmdGen)


class UniqueConstraintCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        if not isinstance(item_data, list):
            item_data = [item_data]
        key_str = ', '.join(item_data)

        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" ADD CONSTRAINT {context['table']}_{item_key} UNIQUE ({key_str})"
        )
        cmd_list.append(cmd_str)
        super().write_add(item_key, item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" DROP CONSTRAINT IF EXISTS {context['table']}_{item_key}"
        )
        cmd_list.append(cmd_str)
        super().write_del(item_key, item_data, cmd_list, context)

    def write_mod(self, item_key, item_data_from, item_data, cmd_list, context):
        self.write_del(item_key, item_data_from, cmd_list, context)
        self.write_add(item_key, item_data, cmd_list, context)
        super().write_mod(item_key, item_data_from, item_data, cmd_list, context)

    def write_dict(self, item_key, item_node, cmd_list, context):
        raise NotImplementedError("write_dict not implemented for UniqueConstraintCmdGen")

CmdGen.register_cmd_gen('unique', 'unique', UniqueConstraintCmdGen)


class ForeignConstraintCmdGen(CmdGen):
    def write_add(self, item_key, item_data, cmd_list, context):
        key_name = f'fk_{item_key}'
        [foreign_table, foreign_col] = item_data.split('.')
        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" ADD CONSTRAINT {key_name}"
            f" FOREIGN KEY ({item_key})"
            f" REFERENCES {context['schema']}.{foreign_table} ({foreign_col})"
        )
        cmd_list.append(cmd_str)
        super().write_add(item_key, item_data, cmd_list, context)

    def write_del(self, item_key, item_data, cmd_list, context):
        key_name = f'fk_{item_key}'
        cmd_str = (
            f"ALTER TABLE {context['schema']}.{context['table']}"
            f" DROP CONSTRAINT IF EXISTS {key_name}"
        )
        cmd_list.append(cmd_str)
        super().write_del(item_key, item_data, cmd_list, context)

    def write_mod(self, item_key, item_data_from, item_data, cmd_list, context):
        self.write_del(item_key, item_data_from, cmd_list, context)
        self.write_add(item_key, item_data, cmd_list, context)
        super().write_mod(item_key, item_data_from, item_data, cmd_list, context)

    def write_dict(self, item_key, item_node, cmd_list, context):
        raise NotImplementedError("write_dict not implemented for ForeignConstraintCmdGen")

CmdGen.register_cmd_gen('foreign', 'foreign', ForeignConstraintCmdGen)


import logging
logger = logging.getLogger(__name__)

import json

from keyrock_core import json_util

from . import exc


def catalog_diff(catalog_from, catalog_to):
    user_types = catalog_to.get('user_types', {})
    if 'schemas' in catalog_to:
        replace_user_types(catalog_to['schemas'], user_types)

    # Drop/Delete these types of items if omitted from structure_to
    explicit_group_list = ['tables', 'columns', 'constraints', 'unique']
    ignore_group_list = ['user_types']

    diff = json_util.diff(catalog_from, catalog_to, explicit_group_list=explicit_group_list, ignore_group_list=ignore_group_list)
    return diff


def replace_user_types(target_dict, user_types):
    # Search for the 'columns' subdict
    for key, val in target_dict.items():
        if key == 'columns':
            replace_user_types_on_columns(val, user_types)
        elif key == 'constraints':
            clean_up_constraints(val)
        else:
            replace_user_types(val, user_types)


def replace_user_types_on_columns(columns_dict, user_types):
    for key, val in columns_dict.items():
        # Substitute the user_type if necessary
        if 'user_type' in val:
            user_type = val['user_type']
            user_type_dict = user_types.get(user_type, {})
            del val['user_type']
            val.update(user_type_dict)

        # Convert array types and type synonyms
        col_type = val['type'].upper()
        col_size = val.get('size')
        if col_type == 'ARRAY':
            val['type'] = f"{val['element_type']}[]"
        elif col_type == 'INT':
            val['type'] = 'INTEGER'
        else:
            val['type'] = col_type

        # @todo:
        # Convert "serial" and "bigserial" types to match db representation
        # col_type = col_to.get('type')
        # if col_type == 'serial':
        #     sequence_name = "'{0}_id_seq'::regclass".format(table_name)
        #     col_to.update({
        #         'type': 'integer',
        #         'default': "nextval({0})".format(sequence_name),
        #     })
        # elif col_type == 'bigserial':
        #     sequence_name = "'{0}_id_seq'::regclass".format(table_name)
        #     col_to.update({
        #         'type': 'bigint',
        #         'default': "nextval({0})".format(sequence_name),
        #     })


def clean_up_constraints(constraints_dict):
    for key, val in constraints_dict.items():
        if key == 'primary':
            if not isinstance(val, list):
                constraints_dict[key] = [val]
        if key == 'unique':
            for uq_key_name, uq_key_cols in val.items():
                if not isinstance(uq_key_cols, list):
                    val[uq_key_name] = [uq_key_cols]
